from munch import Munch
from itertools import chain

from infi.shell import commandgroup, command, render
from infi.shell.types import Str
from infi.shell.utils import strlist
from infi.shell.exceptions import BadArguments

from ...utils import input_password
from ..types import (
    Name,
    ActiveDirectory,
    )
from ..options import QueryOptions
from ..utils import increased_api_timeout
from .. import roles, EntityCommands
from . import ConfigCommandsBase


@commandgroup(parent=ConfigCommandsBase, prefix="ldap")
class LdapConfigCommands(EntityCommands, ConfigCommandsBase):
    ENTITY_TYPE = ActiveDirectory
    QUERY_FIELDS = strlist("""
        name repository_type domain_name
        """)
    DETAILED_QUERY_FIELDS = strlist("""
        name repository_type domain_name bind_username
        user_class username_attribute users_basedn
        group_class group_name_attribute group_memberof_attribute groups_basedn
        """)
    SORTABLE_FIELDS = strlist("name")

    @command(
        description="List existing Active Directory servers",
        types=(ActiveDirectory,),
        substitute=dict(repo="name"),
        options=QueryOptions(fields=QUERY_FIELDS, infinidat_fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS,  detailed=True),
        )
    @render(
        fields=QUERY_FIELDS,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        )
    def query(self, _api, repo=None):
        return self._query(_api, repo, _hook=self._query_hook)

    def _query_hook(cls, results):
        return (cls._object_hook(r) for r in results)

    def _object_hook(cls, repo):
        if "schema_definition" in repo:
            repo.update(repo.schema_definition)
            del repo.schema_definition
        repo.pop("role_mappings", None)
        return repo

    def _create_ldap_params(_self, domain, username, password,
                                 schema_user_class, schema_username_attribute, schema_user_basedn,
                                 schema_group_class, schema_group_attribute, schema_member_attribute, schema_group_basedn, **_):
        return Munch(
            domain_name=domain,
            bind_username=username,
            bind_password=password,
            schema_definition=Munch((k, v) for k, v in dict(
                user_class=schema_user_class,
                username_attribute=schema_username_attribute,
                users_basedn=schema_user_basedn,
                group_class=schema_group_class,
                group_name_attribute=schema_group_attribute,
                group_memberof_attribute=schema_member_attribute,
                groups_basedn=schema_group_basedn,
                ).iteritems() if v is not None),
            )

    @command(
        description="Define Active Directory server",
        types=(Name, Str, Str, Str,
               Str, Str, Str,
               Str, Str, Str, Str),
        )
    @roles.INFINIDAT.ADMIN
    def define(self, _api, name, domain, username, password=None,
                           schema_user_class=None, schema_username_attribute=None, schema_user_basedn=None,
                           schema_group_class=None, schema_group_attribute=None, schema_member_attribute=None, schema_group_basedn=None):
        if password is None:
            password = input_password('Please enter the password for {} "{}": '.format(self.ENTITY_TYPE.get_entity_name(), name),
                                      interaction_exc_type=BadArguments, interaction_error="Please specify a password")
        params = Munch(
            name=name,
            repository_type=ActiveDirectory.REPOSITORY_TYPE,
            )
        params.update(self._create_ldap_params(**locals()))
        with increased_api_timeout():
            return self._create(_api, **params)

    @command(
        description="Modify Active Directory server",
        types=(ActiveDirectory, Str, Str, Str,
               Str, Str, Str,
               Str, Str, Str, Str),
        substitute=dict(repo="name"),
        )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, repo, domain=None, username=None, password=None,
                           schema_user_class=None, schema_username_attribute=None, schema_user_basedn=None,
                           schema_group_class=None, schema_group_attribute=None, schema_member_attribute=None, schema_group_basedn=None):
        if all(arg is None for arg in (
            domain, username, password,
            schema_user_class, schema_username_attribute, schema_user_basedn,
            schema_group_class, schema_group_attribute, schema_member_attribute, schema_group_basedn,
            )):
            raise BadArguments("At least one more argument must be specified")
        if username is not None and password is None:
            password = input_password('Please enter the password for {} "{}": '.format(self.ENTITY_TYPE.get_entity_name(), repo.name),
                                      interaction_exc_type=BadArguments, interaction_error="Please specify a password")
        params = self._create_ldap_params(**locals())
        # Since schema definition must be specified as a whole, it's values are
        # based on the existing definition, and are updated with the specified
        # definition
        params.schema_definition = Munch(chain(
            repo.schema_definition.iteritems(),
            params.schema_definition.iteritems(),
            ))
        with increased_api_timeout():
            return self._update(_api, repo, **params)

    @command(
        description="Rename Active Directory server",
        types=(ActiveDirectory, Name),
        substitute=dict(repo="name"),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, repo, new_name):
        return self._rename(_api, repo, new_name)

    @command(
        description="Undefine Active Directory server",
        types=(ActiveDirectory,),
        substitute=dict(repo="name"),
        )
    @roles.INFINIDAT.ADMIN
    def remove(self, _api, repo):
        return self._delete(_api, repo)

    @command(
        description="Test defined Active Directory server",
        types=(ActiveDirectory,),
        substitute=dict(repo="name"),
        )
    @roles.INFINIDAT.ADMIN
    def test(self, _api, repo):
        self._post(_api, repo, "test")
        return self._result_message(repo, "tested successfully")

    @command(
        description="Set LDAP & Active Directory resolution order",
        types=(ActiveDirectory.List,),
        substitute=dict(repos="order"),
        )
    @roles.INFINIDAT.ADMIN
    def set_order(self, _api, repos):
        self._post(_api, None, "set_order", data=[r.id for r in repos])
        return "Order set"

    @command(
        description="LDAP & Active Directory resolution order",
        types=(),
        substitute=dict(search_order="index"),
        )
    @render(
        fields=strlist('''search_order name'''),
        substitute=dict(search_order="index", name="ldap"),
        )
    def order_query(self, _api):
        return self._query(_api, _hook=self._order_query_hook,_default_sort=_api.create_sort().asc("search_order"))

    def _order_query_hook(cls, results):
        def incr_search_order(x):
            x.search_order += 1
            return x
        return map(incr_search_order, results)

    @command(
        description="Flush LDAP authentication cache",
        )
    @roles.INFINIDAT.ADMIN
    def flush_cache(self, _api):
        self._post(_api, None, "reload")
        return "LDAP authentication cache flushed"
