from itertools import chain, izip

from infi.shell import commandgroup, command, render
from infi.shell.types import (
    Str, YesNo,
    Hostname, Port,
    EmailAddress,
    )
from infi.shell.exceptions import BadArguments
from infi.shell.utils import strlist

from ...utils import input_password
from ..types import (
    Name,
    SmtpTarget,
    SnmpTarget, SnmpProtocol, SnmpSecurity, SnmpAuthProtocol, SnmpEncryption,
    SyslogTarget, SyslogProtocol, SyslogFacility,
    )
from ..options import QueryOptions
from .. import roles, EntityCommands
from . import ConfigCommandsBase

@commandgroup(parent=ConfigCommandsBase, prefix="notifications", manifest=None)
class NotificationCommands(EntityCommands):
    pass

@commandgroup(parent=NotificationCommands, prefix="smtp")
class SmtpNotificationTargetCommands(ConfigCommandsBase, EntityCommands):
    ENTITY_TYPE = SmtpTarget
    QUERY_FIELDS = strlist("name host port tls username from_address subject_prefix")
    SORTABLE_FIELDS = strlist("name host")

    @command(
        description="List existing SMTP notification targets",
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS),
        )
    @render(
        fields=QUERY_FIELDS,
        )
    def query(self, _api):
        return self._query(_api)

    @command(
        description="Modify SMTP notification target",
        types=(SmtpTarget, Hostname, Port, YesNo, Str, Str, EmailAddress, Str),
        substitute=dict(target="name"),
        )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, target, host=None, port=None, tls=None, username=None, password=None, fromaddress=None, subject_prefix=None):
        if username is not None and password is None:
            password = input_password('Please enter the password for {} "{}": '.format(self.ENTITY_TYPE.get_entity_name(), target.name),
                                      interaction_exc_type=BadArguments, interaction_error="Please specify a password")
        return self._update(_api, target,
            host=host,
            port=port,
            tls=tls,
            username=username,
            password=password,
            from_address=fromaddress,
            subject_prefix=subject_prefix,
            )

    @command(
        description="Rename SMTP notification target",
        types=(ENTITY_TYPE, Name),
        substitute=dict(target="name"),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, target, new_name):
        return self._rename(_api, target, new_name)

    @command(
        description="Test SMTP notification target",
        types=(ENTITY_TYPE, EmailAddress),
        substitute=dict(target="name"),
        )
    @roles.INFINIDAT.ADMIN
    def test(self, _api, target, recipient):
        self._post(_api, target, "test", data=dict(recipients=[recipient]))
        return self._result_message(target, "tested successfully")

@commandgroup(parent=NotificationCommands, prefix="snmp")
class SnmpNotificationTargetCommands(ConfigCommandsBase, EntityCommands):
    ENTITY_TYPE = SnmpTarget
    QUERY_FIELDS = strlist("name host port version community engine username auth_type")
    SORTABLE_FIELDS = strlist("name host")
    QUERY_SUBSTITUTES = dict(
        engine="engine_id",
        username="username",
        auth_type="security",
        )

    @command(
        description="List existing SNMP notification targets",
        types=(ENTITY_TYPE,),
        substitute=dict(target="name"),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, target=None):
        return self._query(_api, target)

    def _verify_snmp_1_2_mandatory_args(_self, **kwargs):
        if kwargs["community"] is None:
            raise BadArguments("community must be specified for {} and {}".format(SnmpProtocol.SNMPv1, SnmpProtocol.SNMPv2c))

    def _verify_snmp_1_2_prohibited_args(_self, **kwargs):
        for arg in strlist("engine_id username security password auth_type priv_key encryption"):
            if kwargs[arg] is not None:
                raise BadArguments("{} cannot be specified for {} and {}".format(arg, SnmpProtocol.SNMPv1, SnmpProtocol.SNMPv2c))

    def _verify_snmp_3_prohibited_args(_self, **kwargs):
        if kwargs["community"] is not None:
            raise BadArguments("community must not be specified for {}".format(SnmpProtocol.SNMPv3))

    def _verify_snmp_3_mandatory_args(_self, **kwargs):
        for arg in strlist("engine_id username security"):
            if kwargs[arg] is None:
                raise BadArguments("{} must be specified for {}".format(arg, SnmpProtocol.SNMPv3))

    def _verify_snmp_3_security_args(_self, **kwargs):
        security = kwargs["security"]
        if security == SnmpSecurity.NoAuthNoPriv:
            for arg in strlist("password auth_type priv_key encryption"):
                if kwargs[arg] is not None:
                    raise BadArguments("{} must not be specified for security type {}".format(arg, security))
        if security in (SnmpSecurity.AuthNoPriv, SnmpSecurity.AuthPriv):
            for arg in strlist("auth_type"):
                if kwargs[arg] is None:
                    raise BadArguments("{} must be specified for security type {}".format(arg, security))
            if security == SnmpSecurity.AuthPriv:
                for arg in strlist("priv_key encryption"):
                    if kwargs[arg] is None:
                        raise BadArguments("{} must be specified for security type {}".format(arg, security))

    def _get_snmp_3_password(self, target_name, security, password):
        if password is None and security in (SnmpSecurity.AuthNoPriv, SnmpSecurity.AuthPriv):
            password = input_password('Please enter the password for {} "{}": '.format(self.ENTITY_TYPE.get_entity_name(), target_name),
                                      interaction_exc_type=BadArguments, interaction_error="Please specify a password")
        return password

    @command(
        description="Create a new SNMP notification target",
        types=(Name, Hostname, SnmpProtocol, Port, Str, Str, Str, SnmpSecurity, Str, SnmpAuthProtocol, Str, SnmpEncryption),
        )
    @roles.INFINIDAT.ADMIN
    def define(self, _api, name, host, version, port=None, community=None, engine_id=None, username=None, security=None,
                           password=None, auth_type=None, priv_key=None, encryption=None):
        args = locals()
        params = dict(
            protocol=SnmpTarget.PROTOCOL,
            name=name,
            host=host,
            port=port,
            version=version,
            )
        if version in (SnmpProtocol.SNMPv1, SnmpProtocol.SNMPv2c):
            self._verify_snmp_1_2_prohibited_args(**args)
            self._verify_snmp_1_2_mandatory_args(**args)
            params.update(
                community=community,
                )
        elif version == SnmpProtocol.SNMPv3:
            self._verify_snmp_3_prohibited_args(**args)
            self._verify_snmp_3_mandatory_args(**args)
            self._verify_snmp_3_security_args(**args)
            password = self._get_snmp_3_password(name, security, password)
            params.update(
                engine=engine_id,
                username=username,
                auth_type=security,
                password=password,
                auth_protocol=auth_type,
                private_key=priv_key,
                private_protocol=encryption,
                )
        else:
            raise NotImplementedError("Unsupported SNMP protocol version: {}".format(version))
        return self._create(_api, **params)

    @command(
        description="Modify SNMP notification target",
        types=(SnmpTarget, Hostname, Port, Str, Str, Str, SnmpSecurity, Str, SnmpAuthProtocol, Str, SnmpEncryption),
        substitute=dict(target="name"),
        )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, target, host=None, port=None, community=None, engine_id=None, username=None, security=None,
                                   password=None, auth_type=None, priv_key=None, encryption=None):
        args = locals()
        general_args = (host, port, community, engine_id, username, security)
        security_arg_names = ("password", "auth_type", "priv_key", "encryption")
        security_args = tuple(args[name] for name in security_arg_names)
        if all(arg is None for arg in chain(general_args, security_args)):
            raise BadArguments("At least one more argument must be specified")
        params = dict(  # Fields with None value are ignored
            host=host,
            port=port,
            )
        if target.version in (SnmpProtocol.SNMPv1, SnmpProtocol.SNMPv2c):
            self._verify_snmp_1_2_prohibited_args(**args)
            return self._update(_api, target,
                community=community,
                **params
                )
        elif target.version == SnmpProtocol.SNMPv3:
            self._verify_snmp_3_prohibited_args(**args)
            if not security:
                for arg_name, arg in izip(security_arg_names, security_args):
                    if arg is not None:
                        raise BadArguments("security must be specified along with " + arg_name)
            else:
                self._verify_snmp_3_security_args(**args)
                password = self._get_snmp_3_password(target.name, security, password)
                self._post(_api, target, "update_entity", data=dict(
                    auth_type=security,
                    password=password,
                    auth_protocol=auth_type,
                    private_key=priv_key,
                    private_protocol=encryption,
                    ))
            return self._update(_api, target,
                engine=engine_id,
                username=username,
                **params
                )
        else:
            raise NotImplementedError("Unsupported SNMP protocol version: {}".format(target.version))

    @command(
        description="Rename SNMP notification target",
        types=(ENTITY_TYPE, Name),
        substitute=dict(target="name"),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, target, new_name):
        return self._rename(_api, target, new_name)

    @command(
        description="Delete SNMP notification target",
        types=(ENTITY_TYPE,),
        substitute=dict(target="name"),
        )
    @roles.INFINIDAT.ADMIN
    def remove(self, _api, target):
        return self._delete(_api, target)

    @command(
        description="Test SNMP notification target",
        types=(ENTITY_TYPE,),
        substitute=dict(target="name"),
        )
    @roles.INFINIDAT.ADMIN
    def test(self, _api, target):
        self._post(_api, target, "test", data={})
        return self._result_message(target, "tested successfully")

@commandgroup(parent=NotificationCommands, prefix="syslog")
class SyslogNotificationTargetCommands(ConfigCommandsBase, EntityCommands):
    ENTITY_TYPE = SyslogTarget
    QUERY_FIELDS = strlist("name host port transport facility")
    SORTABLE_FIELDS = strlist("name host")
    QUERY_SUBSTITUTES = dict(
        transport="protocol",
        )

    @command(
        description="List existing " + ENTITY_TYPE.get_entity_plural_name(),
        types=(ENTITY_TYPE,),
        substitute=dict(target="name"),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES
        )
    def query(self, _api, target=None):
        return self._query(_api, target)

    @command(
        description="Create a new " + ENTITY_TYPE.get_entity_name(),
        types=(Name, Hostname, Port, SyslogProtocol, SyslogFacility),
        )
    @roles.INFINIDAT.ADMIN
    def define(self, _api, name, host, port=None, protocol=None, facility=None):
        return self._create(_api,
            protocol=SyslogTarget.PROTOCOL,
            name=name,
            host=host,
            port=port,
            transport=protocol,
            facility=facility,
            )

    @command(
        description="Modify " + ENTITY_TYPE.get_entity_name(),
        types=(SyslogTarget, Hostname, Port, SyslogProtocol, SyslogFacility),
        substitute=dict(target="name")
        )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, target, host=None, port=None, protocol=None, facility=None):
        return self._update(_api, target,
            host=host,
            port=port,
            transport=protocol,
            facility=facility,
            )

    @command(
        description="Rename " + ENTITY_TYPE.get_entity_name(),
        types=(ENTITY_TYPE, Name),
        substitute=dict(target="name"),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, target, new_name):
        return self._rename(_api, target, new_name)

    @command(
        description="Delete " + ENTITY_TYPE.get_entity_name(),
        types=(ENTITY_TYPE,),
        substitute=dict(target="name"),
        )
    @roles.INFINIDAT.ADMIN
    def remove(self, _api, target):
        return self._delete(_api, target)

    @command(
        description="Test " + ENTITY_TYPE.get_entity_name(),
        types=(ENTITY_TYPE,),
        substitute=dict(target="name"),
        )
    @roles.INFINIDAT.ADMIN
    def test(self, _api, target):
        self._post(_api, target, "test", data={})
        return self._result_message(target, "tested successfully")
