from __future__ import print_function, absolute_import

import logging, logging.handlers
import os
import sys
import gzip
import shutil

from infi.shell import config

_CONFIG_ERROR_PREFIX = "Failed to configure internal logging"

def configure_logging(encoding="utf-8"):
    configured = False
    try:
        configured = _configure_logging(encoding)
    except BaseException as e:
        print("{}: {}".format(_CONFIG_ERROR_PREFIX, e), file=sys.stderr)
    if not configured:
        disable_logging()

def _configure_logging(encoding):
    if not config.internal_log_file:
        return False
    if os.path.isfile(config.internal_log_file) and not os.access(config.internal_log_file, os.W_OK):
        print('{}: Log file exists, but is not writable: {}'.format(_CONFIG_ERROR_PREFIX, config.internal_log_file))
        return False
    # Create enclosing directory if it doesn't exist
    basedir = os.path.dirname(config.internal_log_file)
    if not os.path.exists(basedir):
        try:
            os.makedirs(basedir)
        except BaseException:
            print('{}: Failed to create base directory "{}"'.format(_CONFIG_ERROR_PREFIX, basedir), file=sys.stderr)
            return False
    # Configure logger
    root = logging.root
    root.setLevel(logging.DEBUG)
    handler = RotatingGzipFileHandler(
        filename=config.internal_log_file,
        mode="a",
        encoding=encoding,
        maxBytes=config.internal_log_max_size_mb * 1024 * 1024,
        backupCount=config.internal_log_backup_count,
        )
    handler.setFormatter(get_formatter())
    root.addHandler(handler)
    return True

def disable_logging():
    logging.disable(logging.FATAL)

def get_formatter():
    return logging.Formatter(
        fmt="%(asctime)s [%(levelname)5s] (%(name)s) %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        )

class RotatingGzipFileHandler(logging.handlers.RotatingFileHandler):
    """ An extension of :py:class:`logging.handlers.RotatingFileHandler` that
    also compresses the backups. """
    def __init__(self, *args, **kwargs):
        super(RotatingGzipFileHandler, self).__init__(*args, **kwargs)
        self.file_template = "{{}}.{{:0{}}}.gz".format(len(str(self.backupCount)))
    def doRollover(self):
        if self.stream:
            self.stream.close()
            self.stream = None
        if self.backupCount > 0:
            for i in range(self.backupCount - 1, 0, -1):
                src_file = self.file_template.format(self.baseFilename, i)
                dest_file = self.file_template.format(self.baseFilename, i + 1)
                if os.path.exists(src_file):
                    if os.path.exists(dest_file):
                        os.remove(dest_file)
                    os.rename(src_file, dest_file)
            dest_file = self.file_template.format(self.baseFilename, 1)
            if os.path.exists(dest_file):
                os.remove(dest_file)
            with gzip.open(dest_file, "wb") as fdest:
                with open(self.baseFilename) as fsrc:
                    shutil.copyfileobj(fsrc, fdest, length=4*1024*1024)
            os.remove(self.baseFilename)
        self.stream = self._open()
