from __future__ import print_function

# Import patches first
from . import patches  # noqa
import infi.shell.utils.colorama_patches  # noqa

# Import everything else
import sys
import traceback
import time
import socket
import itertools
import textwrap
from optparse import OptionParser, IndentedHelpFormatter
from collections import namedtuple
from getpass import getpass
from functools import partial

from infi.shell import (
    Commands,
    Parser, ShellParser,
    config, default_config, write_default_config, initialize_config,
    datadir,
    )
from infi.shell.parser import CHARS as PARSER_CHARS
from infi.shell.exceptions import FilesystemException
from infi.shell.formatters import format_exception

from . import debug
from . import raw_mode
from . import VERSION
from . import infinishell_config # Import explicitly to apply Infinibox configuration # noqa
from .logging import configure_logging
from .utils import post_mortem
from .shell import InfiniboxShell, InfiniboxExecutor
from .api import InfiniboxAPI
from .args_processor import InfiniboxArgsProcessor
from .environment import variables as env
from .formatters import InfinidatExceptionFormatter
# Due to cyclic dependencies, types has to be imported before renderers
from .commands import types as ___  # noqa
from .commands.options import DebugOption, AutoApprove
from .renderer import InfiniboxRenderer, toggle_csv_mode, toggle_json_mode
from .plugins import plugins
from .system.system import System, system as system_context
from .system.user import Roles

from .constants import (
    COMPANY_NAME,
    COMPANY_NAME_LOWER,
    PRODUCT_NAME,
    CLIENT_NAME,
    CLIENT_NAME_LOWER,
    CLIENT_NAME_UPPER,
    ROOT_USER,
    )

from .exceptions import (
    InvalidOptions,
    ConnectionFailed,
    ServiceTemporarilyUnavailable,
    InternalServerError,
    ManagementException,
    UnexpectedServerResponse,
    APIFailure,
    )

# Import command modules so that they register in the manifest
from .manifest import manifest

from .pre_v2_patches import pre_v2_patches

env.set_prefix(CLIENT_NAME_UPPER)

WELCOME_MSG = "{} Ltd. Command Line Utility v{}".format(COMPANY_NAME, VERSION)
WELCOME_BANNER = "\n{}\n".format(WELCOME_MSG)
EXIT_MSG = "Goodbye!"

BASE_URI = "api"
REST_BASE_URI = "rest"
INTERNAL_BASE_URI = "internal"

DATA_DIR = "~/.{}/{}".format(COMPANY_NAME_LOWER, CLIENT_NAME_LOWER)
CONFIG_FILE = CLIENT_NAME_LOWER + "rc"


ENCODING = "utf-8"
COLOR_OPTIONS = ["always", "never", "auto"]
NO_VALUE = "__no_value__"

COMMANDS = [cls() for cls in manifest]
GLOBAL_OPTIONS = (
    DebugOption(),
    AutoApprove(),
    )

FlagOption = namedtuple("FlagOption", ("name", "enabler_option", "disabler_option"))

def output(msg, *args, **kwargs):
    print(msg.format(*args, **kwargs), file=sys.stdout)

def error(msg, *args, **kwargs):
    prefix = "ERROR: "
    prefixes = itertools.chain([prefix], itertools.cycle([" "*len(prefix)]))
    for line in textwrap.wrap(msg.format(*args, **kwargs), 70-len(prefix)):
        print(next(prefixes) + line, file=sys.stderr)

class InfiniShellOptionParser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)
        self._flag_options = []
    def error(self, msg):
        self.print_usage(sys.stderr)
        raise InvalidOptions(msg.capitalize())
    def add_flag_option(self, flag, enabler_options, enabler_help, disabler_options, disabler_help):
        default_value = getattr(default_config, flag)
        enabler_dest, disabler_dest = self._get_dest_name(enabler_options), self._get_dest_name(disabler_options)
        for options, help, dest, xor in ((enabler_options, enabler_help, enabler_dest, False),
                                         (disabler_options, disabler_help, disabler_dest, True)):
            is_enabled = default_value^xor
            self.add_option(*options, dest=dest, action="store_true", default=False,
                            help="{} (this is the default)".format(help) if is_enabled else help)
        self._flag_options.append(FlagOption(flag, enabler_dest, disabler_dest))
    def add_config_option(self, *args, **kwargs):
        assert args, "Option names are mandatory"
        assert "default" not in kwargs, "Defaults are not allowed in config options"
        key = kwargs.get("dest")
        key = key or kwargs.setdefault("dest", args[-1][2:] if args[-1].startswith("--") else args[-1][1:])
        if "help" in kwargs:
            kwargs["help"] = kwargs["help"].replace("%default", "{}").format(default_config[key])
        kwargs.setdefault("type", "string")
        if kwargs["type"] == "string":
            kwargs["default"] = NO_VALUE
        kwargs.update(
                action="callback",
                callback=self._config_option_callback,
                )
        self.add_option(*args, **kwargs)
    def _config_option_callback(self, option, option_str, value, parser):
        assert option.dest
        config[option.dest] = value
        setattr(parser.values, option.dest, value)
    def _get_dest_name(self, options):
        returned = None
        for option in options:
            if option[0:2] == "--":
                returned = option[2:]
                break
        return (returned or options[0][1:]).replace("-", "_")
    def set_options_in_config(self, values):
        for flag in self._flag_options:
            is_enabled = config[flag.name]
            if (is_enabled and getattr(values, flag.disabler_option) or
                not is_enabled and getattr(values, flag.enabler_option)):
                config[flag.name] = not is_enabled

def display_version_and_exit(*_, **__):  # noqa
    output(WELCOME_MSG)
    sys.exit(0)

def write_default_config_and_exit(*_, **__): # noqa
    if datadir.exists(CONFIG_FILE):
        error("Configuration file already exists. Delete it first.")
    else:
        try:
            config_file_path = datadir.get_path(CONFIG_FILE)
            write_default_config(config_file_path)
            output('Default configuration file generated at "{}"'.format(config_file_path))
            sys.exit(0)
        except OSError as e:
            error("Failed to write configuration file: {}", e)
    sys.exit(2)

def list_env_vars_and_exit(*_, **__):
    def var_to_str(v):
        returned = "{}_{}".format(env.get_prefix(), v.name)
        if v.is_flag:
            returned += " (flag)"
        if v.help:
            returned += ": " + v.help
        if v.default is not None:
            returned += " (default: {})".format(v.default)
        return returned
    env_vars = "\n  ".join(var_to_str(v) for v in env.get_all() if not v.is_internal)
    output('Environment variables:\n  ' + env_vars)
    output('\nPossible flag values: ' + ', '.join(itertools.chain(env.Variable.POSITIVES,
                                                                env.Variable.NEGATIVES)))
    sys.exit(0)

def assert_options(options, args):
    if len(args) > 1:
        raise InvalidOptions("Too many arguments: {}".format(", ".join(repr(arg) for arg in args)), print_usage=True)
    if (options.command, options.file, options.command_help).count(None) < 2:
        raise InvalidOptions("Please specify at most one of the following arguments: -c, -f, -H")
    if options.color not in ([NO_VALUE] + COLOR_OPTIONS):
        raise InvalidOptions("Options for --color are: {}".format(", ".join(COLOR_OPTIONS)))

def handle_debug(options):
    if options.debug:
        debug.on()

def handle_raw_mode(options):
    if options.raw:
        raw_mode.toggle()

def get_script(options):
    if options.command:
        script = options.command
        if PARSER_CHARS.cmd_separator in script:
            raise InvalidOptions("Multiple commands are only allowed in scripts")
    elif isinstance(options.command, basestring):
        raise InvalidOptions("No command was specified")
    elif options.file:
        with open(options.file) as f:
            script = f.read()
    elif options.command_help:
        script = "help on='{}'".format(options.command_help)
    elif not sys.stdin.isatty():
        script = sys.stdin.read()
    else:
        return None
    return script.decode(ENCODING)

def get_address(args):
    address = args[0] if args else env.get("HOSTNAME", config.default_address)
    if not address:
        raise InvalidOptions("Please specify system address", print_usage=True)
    address_port = address.split(":")
    address = address_port[0]
    if len(address_port) == 1:
        port = None
    elif len(address_port) == 2:
        port = int(address_port[1])
    else:
        raise InvalidOptions("Invalid address: {}".format(address))
    return address, port

def simplify_address(address):
    address = address.lower()
    simplified_address = "".join(c for c in address if c in ("abcdefghijklmnopqrstuvwxyz1234567890-_"))
    return "{}_{:x}".format(simplified_address, abs(hash(address)))

def get_history_filename(address):
    if not config.history:
        return None
    if config.history_file:
        return config.history_file
    simplified_address = simplify_address(address)
    return datadir.get_path("history", simplified_address)

def new_client_id():
    localhost = socket.gethostname()
    return "{}-{}-{:x}".format(CLIENT_NAME, localhost, int(time.time()*1000))


def check_connection(system, address):
    output("Connecting to {}...", address)
    system.api.get("system")
    system.user

def set_username_password(options):
    if env.is_flagged("NO_AUTH"):
        return
    if options.password is not NO_VALUE and options.username is NO_VALUE:
        raise InvalidOptions("Please specify either a username, a username and password, or neither.")
    def set_password():
        config.password = getpass("{} password: ".format(PRODUCT_NAME))
    if not config.username:
        auth = env.get("AUTH")
        if auth:
            auth = auth.split(":", 1)
            if len(auth) != 2:
                raise InvalidOptions('{}_AUTH environment variable should contain '
                     'colon-separated username and password, e.g: "user:pass"'.format(CLIENT_NAME_UPPER))

            config.username, config.password = auth
        else:
            if not sys.stdin.isatty():
                raise InvalidOptions("Credentials not set, and cannot get them while running "
                                     "non-interactively. Use the configuration file (see help), "
                                     "or {}_AUTH environment variable.".format(CLIENT_NAME_UPPER))

            config.username = raw_input("{} username: ".format(PRODUCT_NAME))
            set_password()
    elif not config.password:
        if not sys.stdin.isatty():
            raise InvalidOptions("Only username was set in configuration file, and "
                                 "cannot get password while running non-interactively.")
        set_password()

def get_default_headers():
    headers = dict()
    for header, var in (
            ("X-Forwarded-For", "X_FORWARDED_FOR"),
            ):
        if env.is_set(var):
            headers[header] = env.get(var)
    return headers

def prompt_hook(shell, address, system):
    return "{}@{}".format(system.user.username, address)

def run(address, port, script=None):
    try:
        try:
            is_interactive = script is None
            ssl_enabled = config.api_ssl and not env.is_flagged("NO_SSL")
            base_api = InfiniboxAPI(
                    "{protocol}://{address}:{port}/{base_uri}".format(
                        protocol="https" if ssl_enabled else "http",
                        address=address,
                        port=port or (config.api_ssl_port if ssl_enabled else config.api_port),
                        base_uri=BASE_URI,
                        ),
                    new_client_id(),
                    headers=get_default_headers(),
                    encoding=ENCODING,
                    )
            api = base_api.get_sub_resource(REST_BASE_URI, internal=base_api.get_sub_resource(INTERNAL_BASE_URI))

            system = System(api)
            system_context.__push_context__(system)


            if is_interactive:
                output("Connecting to {}...", address)

            # The no-auth mode comes to work around having to authenticate when
            # using the shell from within the system's OS shell (SSH)
            if env.is_flagged("NO_AUTH"):
                system._set_user(ROOT_USER, [Roles.INFINIDAT])
            elif env.is_set("EMULATE_LOGIN"):
                login_info = env.get("EMULATE_LOGIN").split(",", 2)
                if len(login_info) != 3:
                    raise InvalidOptions('Environment variable {}_EMULATE_LOGIN format is invalid: "{}".\n'
                                         'Expected format: "username,password,role[,role,...]"'
                                         .format(env.get_prefix(), login_info))
                [username, password, roles] = login_info
                system._set_user(username, roles.split(","))
                api.set_basic_auth(username, password)
            else:
                system.login(config.username, config.password)  # Also sets API basic auth

            pre_v2_patches(api)

            commands = Commands(
                    *(list(COMMANDS) + list(plugins.create_commands())),
                    global_options=GLOBAL_OPTIONS,
                    args_processor=InfiniboxArgsProcessor(),
                    default_renderer=InfiniboxRenderer(),
                    api=api
                )

            parser_kwargs = dict(commands=commands, encoding=ENCODING)
            _color_str = config.color.lower()
            _color = True if _color_str == "always" else False if _color_str == "never" else None
            out_stream = sys.stdout
            common_kwargs = dict(
                stream=out_stream,
                paging=config.paging,
                exc_formatter=InfinidatExceptionFormatter(),
                color=_color,
                errors_color=_color,
                )

            def prepare(shell):
                if config.log_file:
                    shell.start_logging(config.log_file)
                if config.csv_output:
                    toggle_csv_mode(shell)
                if config.json_output:
                    toggle_json_mode(shell)

            result_code = 0

            if is_interactive:
                shell = InfiniboxShell(
                    parser=ShellParser(**parser_kwargs),
                    history_filename=get_history_filename(address),
                    encoding=ENCODING,
                    prompt=partial(prompt_hook, address=address, system=system),
                    welcome_msg=WELCOME_BANNER,
                    exit_msg=EXIT_MSG,
                    error_stream=out_stream,  # In interactive mode stderr is not used
                    **common_kwargs
                    )
            else:
                shell = InfiniboxExecutor(
                    parser=Parser(**parser_kwargs),
                    keep_going=config.keep_going,
                    quiet=config.quiet,
                    **common_kwargs
                    )

            prepare(shell)
            plugins.before_any_session(shell)

            if is_interactive:
                plugins.before_interactive_session(shell)
                shell.start()
                plugins.after_interactive_session(shell)
            else:
                plugins.before_script_session(shell)
                success = shell.execute(script)
                result_code = 0 if success else 2
                plugins.after_script_session(shell, success)

            plugins.after_any_session(shell)

            return result_code
        except:
            # In debug mode, stack trace should be printed for all exceptions
            if env.is_flagged("DEBUG"):
                traceback.print_exc()
            raise
    except (ConnectionFailed, ManagementException, UnexpectedServerResponse) as e:
        error("Unable to reach {}: {}", PRODUCT_NAME, str(e))
        return 2
    except ServiceTemporarilyUnavailable:
        error("{} temporarily unavailable", PRODUCT_NAME)
        return 2
    except InternalServerError:
        error("An unknown error occurred on the server side: {}", format_exception())
        return 2
    except APIFailure as e:
        msg = "Failed to communicate with server due to product or version mismatch"
        if env.is_flagged("DEBUG") or debug.is_on():
            error(msg + ": {}", format_exception())
        else:
            error(msg)
        return 2
    except FilesystemException as e:
        error(str(e))
        return 1
    except (InvalidOptions, KeyboardInterrupt):
        raise
    except:
        # Save exc_info now, in case other exceptions arise before post mortem
        exc_info = sys.exc_info()
        # For unknown exceptions, stack trace should be printed
        if not env.is_flagged("DEBUG"):
            if debug.is_on():
                traceback.print_exc()
            error("An unexpected error has occurred - please contact customer support")
        # In debug mode, perform post mortem
        else:
            post_mortem(exc_info)
        return 1

def apply_plugins_config():
    for defaults in plugins.get_config():
        if defaults:
            default_config.update(defaults)

def _main():
    datadir.init(DATA_DIR)
    config_file_path = datadir.get_path(CONFIG_FILE)
    apply_plugins_config()
    initialize_config(config_file_path)
    op = InfiniShellOptionParser(prog=CLIENT_NAME_LOWER, usage="Usage: %prog [options] system[:port]",
                                 epilog='Please note that defaults may be overridden in the configuration '
                                        'file: "{}". Also, some environment variables are available for '
                                        'your convenience. To list them, use the --list-environment option.'
                                        .format(config_file_path),
                                 formatter=IndentedHelpFormatter(max_help_position=30))
    op.add_option("-v", "--version", action="callback", callback=display_version_and_exit,
                  help="Print version and exit")
    op.add_option("-c", "--cmd", dest="command", metavar="CMD", default=None,
                  help="Commands to execute (separated by semicolons)")
    op.add_option("-f", "--file", metavar="FILE", default=None,
                  help="Script file to execute")
    op.add_option("-d", "--debug", action="store_true", default=False,
                  help="Enable debug prints")
    op.add_option("-r", "--raw", action="store_true", default=False,
                  help="Display raw API responses instead of normal rendering")
    op.add_config_option("-u", "--user", dest="username",
                         help="Login username")
    op.add_config_option("-p", "--password",
                         help="Login password")
    op.add_option("-H", dest="command_help", metavar="CMD", default=None,
                  help="Display syntax of the specified command")
    op.add_config_option("-l", "--log", dest="log_file", metavar="FILE",
                         help="Log commands to the specified file (default: %default)")
    op.add_option("--write-default-config", action="callback", callback=write_default_config_and_exit,
                  help='Generate the default configuration file at "{}"'.format(config_file_path))
    op.add_option("--list-environment", action="callback", callback=list_env_vars_and_exit,
                  help="List environment variables")
    op.add_config_option("--color", metavar="WHEN",
                         help="Whether or not to output with color. Options are: always, never, auto (default: %default)")
    op.add_flag_option(flag="keep_going",
                       enabler_options=("-k", "--keep-going"),
                       enabler_help="Keep going when administration errors occur (syntax errors always abort the entire script)",
                       disabler_options=("--fail-fast",),
                       disabler_help="Stop running script when the first administration error occurs")
    op.add_flag_option(flag="quiet",
                       enabler_options=("-q", "--quiet"),
                       enabler_help="No output apart from errors",
                       disabler_options=("--not-quiet",),
                       disabler_help="Normal output, the opposite of --quiet")
    op.add_flag_option(flag="paging",
                       enabler_options=("--with-paging",),
                       enabler_help="Enable paging of large outputs",
                       disabler_options=("--no-paging",),
                       disabler_help="Disable paging of large outputs")
    op.add_flag_option(flag="auto_confirm_prompts",
                       enabler_options=("-y", "--yes"),
                       enabler_help="Automatically confirm prompts (of potentially destructive commands)",
                       disabler_options=("--no",),
                       disabler_help="Automatically reject prompts (of potentially destructive commands)")
    op.add_config_option("--history-file", metavar="PATH", dest="history_file",
                         help="Path of history file, for cross-session history of commands. By default this "
                              "is saved in {}'s data directory, under a unique name generated off of "
                              "the target system address".format(CLIENT_NAME))
    op.add_flag_option(flag="history",
                       enabler_options=("--with-history",),
                       enabler_help="Enable cross-session history of commands",
                       disabler_options=("--no-history",),
                       disabler_help="Disable cross-session history of commands")
    op.add_flag_option(flag="csv_output",
                       enabler_options=("--csv",),
                       enabler_help="CSV output mode (optimize output for easy parsing, see RFC 4180)",
                       disabler_options=("--no-csv",),
                       disabler_help="Disable CSV output mode")
    op.add_flag_option(flag="json_output",
                       enabler_options=("--json",),
                       enabler_help="JSON output mode (optimize output for easy parsing, see RFC 4180)",
                       disabler_options=("--no-json",),
                       disabler_help="Disable JSON output mode")
    op.add_config_option("--request-timeout", dest="api_request_timeout_seconds", metavar="SECONDS",
                         type="int",
                         help="Maximal amount of time to wait for management response (default: %default)")
    op.add_flag_option(flag="api_ssl",
                       enabler_options=("--with-ssl",),
                       enabler_help="Encrypt connection using SSL or TLS",
                       disabler_options=("--no-ssl",),
                       disabler_help="Do not encrypt connection")
    try:
        options, args = op.parse_args()
        assert_options(options, args)
        handle_debug(options)
        handle_raw_mode(options)
        op.set_options_in_config(options)
        address, port = get_address(args)
        script = get_script(options)
        set_username_password(options)
        configure_logging(encoding=ENCODING)
        return run(address, port, script)
    except KeyboardInterrupt:
        error("Aborted")
        return 1
    except Exception as e:
        if isinstance(e, InvalidOptions):
            if e.print_help:
                op.print_help()
            elif e.print_usage:
                op.print_usage()
        if env.is_flagged("DEBUG"):
            traceback.print_exc()
        else:
            error(str(e))
        return 1

def main():
    if env.is_flagged("PROFILER"):
        import cProfile
        cProfile.run("from infinishell.infinibox.main import _main; _main()", sort="cumulative")
        return 0
    else:
        return _main()

if __name__ == '__main__':
    sys.exit(main())

