from __future__ import print_function

import sys
import pkg_resources
import logging
import os
import imp
import traceback
from inspect import isroutine, isclass, ismodule
from itertools import chain

from infi.shell import config
from infi.shell.utils import datadir

# Setuptools plugins
ENTRY_POINT_GROUP = "infinishell.infinibox.plugins"

# Filesystem plugins
PLUGINS_SUBDIR = "plugins"
PLUGIN_FILE_NAME = "plugin.py"
PLUGIN_BASE_PACKAGE = "infinishell.infinibox.plugins"

logger = logging.getLogger(PLUGIN_BASE_PACKAGE)
_loaded_containers = None

def _error(msg, *fargs, **fkwargs):
    exc_info = fkwargs.pop("exc_info", False)
    if fargs or fkwargs:
        msg = msg.format(*fargs, **fkwargs)
    print("ERROR: ", end="", file=sys.stderr)
    print(msg, file=sys.stderr)
    if exc_info:
        traceback.print_exc()
    logger.error(msg, exc_info=exc_info)

def load():
    global _loaded_containers
    _loaded_containers = list(chain(
        _load_setuptools_plugins(),
        _load_filesystem_plugins(),
        ))

def _load_setuptools_plugins():
    logger.debug("Searching for plugins through setuptools")
    for ep in pkg_resources.iter_entry_points(group=ENTRY_POINT_GROUP):
        plugin_name, plugin_dist = ep.name, ep.dist.egg_name()
        plugin_fullname = '"{}" ({})'.format(plugin_name, plugin_dist)
        # Load plugin from entry point
        try:
            logger.debug("Loading plugin through setuptools: %s", plugin_fullname)
            ep_target = ep.load()
        except:
            _error("Failed to load plugin {}: exception occurred", plugin_fullname, exc_info=True)
        if ismodule(ep_target):
            for plugin in _load_module_plugins(ep_target):
                yield plugin
        elif isclass(ep_target):
            if not issubclass(ep_target, Plugin):
                _error("Failed to load plugin {}: plugin class does not inherit from {}.{}",
                       plugin_fullname, Plugin.__module__, Plugin.__name__)
                continue
            # Instantiate plugin
            try:
                plugin = ep_target()
            except:
                _error("Failed to load plugin {}: failed to instantiate plugin class", exc_info=True)
                continue
            # Add plugin to list of plugins
            yield PluginContainer(plugin_name, plugin_dist, plugin)
        else:
            _error("Failed to load plugin {}: entry point must either point to a subclass of {}.{}, or to a module, but instead points to a {}",
                   plugin_fullname, Plugin.__module__, Plugin.__name__, type(ep_target))

def _load_filesystem_plugins():
    for plugins_root in (
        config.plugins_dir,
        datadir.get_path(PLUGINS_SUBDIR),
        ):
        for plugin in _load_filesystem_plugins_from(plugins_root):
            yield plugin

def _load_filesystem_plugins_from(plugins_root):
    if not (plugins_root and os.path.isdir(plugins_root)):
        return
    logger.debug("Searching for plugins in %s", plugins_root)
    for plugin_dir in os.listdir(plugins_root):
        if plugin_dir[0] == ".":
            continue
        plugin_name, plugin_dir = plugin_dir, os.path.join(plugins_root, plugin_dir)
        plugin_dist = plugin_dir
        if not os.path.isdir(plugin_dir):
            continue
        plugin_fullname = '"{}" ({})'.format(plugin_name, plugin_dist)
        plugin_file = os.path.join(plugin_dir, PLUGIN_FILE_NAME)
        # Validate plugin file exists
        if not os.path.isfile(plugin_file):
            _error("Failed to load plugin {}: file not found: {}", plugin_fullname, plugin_file)
            continue
        # Load plugin module
        try:
            logger.debug("Loading plugin: %s", plugin_fullname)
            plugin_module = imp.load_source("{}.{}".format(PLUGIN_BASE_PACKAGE, plugin_name), plugin_file)
        except:
            _error("Failed to load plugin {}: exception occurred", plugin_fullname, exc_info=True)
            continue
        for plugin in _load_module_plugins(plugin_name, plugin_dist, plugin_fullname, plugin_module):
            yield plugin

def _load_module_plugins(plugin_name, plugin_dist, plugin_fullname, plugin_module):
    logger.debug("Loading plugins from module %s", plugin_module)
    # Find actual plugins in module
    plugins_found = False
    for obj in plugin_module.__dict__.itervalues():
        if isclass(obj) and issubclass(obj, Plugin):
            plugins_found = True
            # Instantiate plugin
            try:
                plugin = obj()
            except:
                _error("Failed to load plugin {}: failed to instantiate plugin class", plugin_fullname, exc_info=True)
                continue
            # Add plugin to list of plugins
            yield PluginContainer(plugin_name, plugin_dist, plugin)
    if not plugins_found:
        _error("Failed to load plugin {}: no subclasses of {}.{} were found in {}", plugin_fullname, Plugin.__module__, Plugin.__name__, PLUGIN_FILE_NAME)

def iter_plugin_containers():
    global _loaded_containers
    if _loaded_containers is None:
        load()
    return iter(_loaded_containers)

class Plugin(object):
    # Plugin methods
    def before_any_session(self, shell):
        pass
    def after_any_session(self, shell):
        pass
    def before_interactive_session(self, shell):
        pass
    def after_interactive_session(self, shell):
        pass
    def before_script_session(self, shell):
        pass
    def after_script_session(self, shell, success):
        pass
    def get_command_group_classes(self):
        pass
    def get_config(self):
        pass
    # Utility methods
    @property
    def logger(self):
        logger = getattr(self, "__plugin_logger__", None)
        if not logger:
            logger = self.__plugin_logger__ = logging.getLogger("{}.{}".format(PLUGIN_BASE_PACKAGE, self.__plugin_container__.name))
        return logger

class PluginContainer(object):
    def __init__(self, name, dist, plugin):
        self.name = name
        self.dist = dist
        self.plugin = plugin
        plugin.__plugin_container__ = self

class AllPluginsProxy(object):
    _METHOD_NAMES = [a for a in dir(Plugin) if a[0] != "_" and isroutine(getattr(Plugin, a))]
    def __getattr__(self, attr):
        if attr not in self._METHOD_NAMES:
            raise AttributeError("Plugins proxy has no such attribute: " + attr)
        return self.PluginMethodProxy(attr)
    class PluginMethodProxy(object):
        def __init__(self, method_name):
            self._method_name = method_name
        def __call__(self, *args, **kwargs):
            results = []
            for container in iter_plugin_containers():
                if not hasattr(container.plugin, self._method_name):
                    _error('Plugin "{}" lacks method "{}"', container.name)
                method = getattr(container.plugin, self._method_name)
                result = None
                try:
                    result = method(*args, **kwargs)
                except BaseException:
                    _error('Plugin "{}" raised exception from method "{}"', container.name, self._method_name, exc_info=True)
                results.append(result)
            return results
    def create_commands(self):
        for container in iter_plugin_containers():
            try:
                classes = container.plugin.get_command_group_classes()
            except BaseException:
                _error('Plugin "{}" raised exception from method "get_command_group_classes"', container.name, exc_info=True)
            if not classes:
                continue
            if not hasattr(classes, "__iter__"):
                _error('Plugin "{}" method "{}" should return an iterable, but returned "{!r}" instead', container.name, "get_command_group_classes", classes)
            for cls in classes:
                if not isinstance(cls, type):
                    _error('Plugin "{}" method "{}" returned an iterable, but one of its elements is not a class', container.name, "get_command_group_classes")
                    continue
                try:
                    yield cls()
                except BaseException:
                    _error('Plugin "{}" command group instantiation failed', container.name, exc_info=True)

plugins = AllPluginsProxy()
del AllPluginsProxy
